import pkg_resources

from PyQt5 import uic, QtWidgets

from . import export


class ExportDialog(QtWidgets.QDialog):
    _instance_counter = 0

    def __init__(self, parent, fdist_list, identifier, *args, **kwargs):
        """Base class for force-indentation analysis"""
        super(ExportDialog, self).__init__(parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "dlg_export_vals.ui")
        uic.loadUi(path_ui, self)

        self.fdist_list = fdist_list
        self.identifier = identifier

    def done(self, r):
        if r:
            fname, _e = QtWidgets.QFileDialog.getSaveFileName(
                self.parent(),
                "Save metadata and results",
                "pyjibe_export_{:03d}.tsv".format(self.identifier),
                "Tab Separated Values (*.tsv)"
            )

            if fname:
                if not fname.endswith(".tsv"):
                    fname += ".tsv"
                user_choices = {
                    "acquisition": self.checkBox_acquisition,
                    "dataset": self.checkBox_dataset,
                    "qmap": self.checkBox_qmap,
                    "setup": self.checkBox_setup,
                    "storage": self.checkBox_storage,
                    "params_initial": self.checkBox_initial,
                    "params_fitted": self.checkBox_fitted,
                    "params_ancillary": self.checkBox_ancillary,
                    "rating": self.checkBox_rating,
                }
                which = []
                for mdat in user_choices:
                    if user_choices[mdat].isChecked():
                        which.append(mdat)

                export.save_tsv_metadata_results(filename=fname,
                                                 fdist_list=self.fdist_list,
                                                 which=which)
        super(ExportDialog, self).done(r)



"""Test of data set functionalities"""
import pathlib

import numpy as np
from PyQt5 import QtWidgets
import pytest

import pyjibe.head


here = pathlib.Path(__file__).parent
jpkfile = here / "data" / "spot3-0192.jpk-force"


def cleanup_autosave(jpkfile):
    """Remove autosave files"""
    path = jpkfile.parent
    files = path.glob("*.tsv")
    files = [f for f in files if f.name.startswith("pyjibe_")]
    [f.unlink() for f in files]


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    cleanup_autosave(jpkfile)
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    cleanup_autosave(jpkfile)


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = pyjibe.head.PyJibe()
    main_window.close()


def test_clear_and_verify_data(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile])
    war = main_window.subwindows[0].widget()
    # clear data
    tpp = war.tab_preprocess
    tpp.list_preproc_applied.clear()
    war.cb_autosave.setChecked(0)
    # perform simple filter
    item = QtWidgets.QListWidgetItem()
    item.setText("compute_tip_position")
    tpp.list_preproc_applied.addItem(item)
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # disable weighting
    war.tab_fit.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    apret = war.data_set[0]
    war.tab_fit.fit_approach_retract(apret)
    # Now check something
    ftab = war.tab_fit.table_parameters_fitted
    # E
    assert np.allclose(apret.fit_properties["params_fitted"]["E"].value,
                       14741.958242422093,
                       atol=1e-4,
                       rtol=0)
    assert float(ftab.item(0, 0).text()) == 14742
    # contact_point
    assert float(ftab.item(1, 0).text()) == 18029
    # baseline_offset
    assert float(ftab.item(2, 0).text()) == -480.67


def test_fit_all(qtbot):
    """Perform a simple fit with the standard parameters"""
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile, jpkfile])
    war = main_window.subwindows[0].widget()
    war.cb_autosave.setChecked(0)
    war.on_fit_all()
    a1 = war.data_set[0]
    a2 = war.data_set[1]
    assert a1.fit_properties == a2.fit_properties



import numpy as np

import pyjibe.fd.tab_info


def test_all_nan_dict():
    assert pyjibe.fd.tab_info.has_all_nans({"a": np.nan, "b": np.nan})
    assert not pyjibe.fd.tab_info.has_all_nans({"a": 1, "b": np.nan})
    assert not pyjibe.fd.tab_info.has_all_nans({"a": "1", "b": np.nan})



"""Settings file management"""
import numbers
import pathlib

import appdirs

#: application name
NAME = "PyJibe"
#: application author
ORG = "AFM-Analysis"

#: default configuration parameters
DEFAULTS = {}


class SettingsFile(object):
    """Manages a settings file in the user's config dir"""

    def __init__(self, name=NAME, org=ORG, defaults=DEFAULTS, directory=None):
        """Initialize settings file (create if it does not exist)"""
        if directory is None:
            directory = appdirs.user_config_dir(appname=name, appauthor=org)
        fname = pathlib.Path(directory) / (name.replace(" ", "_") + ".cfg")
        # create file if not existent
        if not fname.exists():
            fname.parent.mkdir(exist_ok=True, parents=True)
            # Create the file
            fname.open("w").close()

        self.cfgfile = fname
        self.defaults = defaults
        self.working_directories = {}

    def _get_value(self, dtype, key):
        key = key.lower()
        cdict = self.load()
        if key in cdict:
            val = cdict[key]
        elif dtype in self.defaults and key in self.defaults[dtype]:
            val = self.defaults[dtype][key]
        else:
            raise KeyError(
                "Config key '{}' of type '{}' not set!".format(key, dtype))

        if dtype == "bool":
            if isinstance(val, bool):
                pass
            else:
                if val.lower() not in ["true", "false"]:
                    raise ValueError(
                        "Config key '{}' not boolean!".format(key))
                elif val.lower() == "true":
                    val = True
                else:
                    val = False
        elif dtype == "integer":
            if isinstance(val, numbers.Integral):
                pass
            else:
                if not val.isdigit():
                    raise ValueError(
                        "Config key '{}' is no integer!".format(key))
                val = int(val)
        elif dtype == "string_list":
            if isinstance(val, list):
                pass
            else:
                val = val.split("\t")

        return val

    def get_bool(self, key):
        """Returns boolean configuration key"""
        return self._get_value("bool", key)

    def get_int(self, key):
        """Returns integer configuration key"""
        return self._get_value("integer", key)

    def get_path(self, name=""):
        """Returns the path for label `name`"""
        wdkey = "path {}".format(name.lower())
        try:
            path = self._get_value("path", wdkey)
        except KeyError:
            path = "."
        return path

    def get_string(self, key):
        """Return string"""
        try:
            val = self._get_value("string", key)
        except KeyError:
            val = ""
        return val

    def get_string_list(self, key):
        """Returns list of strings"""
        return self._get_value("string_list", key)

    def load(self):
        """Loads the settings file returning a dictionary"""
        with self.cfgfile.open() as fop:
            fc = fop.readlines()
        cdict = {}
        for line in fc:
            line = line.strip()
            var, val = line.split("=", 1)
            cdict[var.lower().strip()] = val.strip()
        return cdict

    def save(self, cdict):
        """Save a settings dictionary into a file"""
        if not self.cfgfile:
            raise SettingsFileError("Settings path not set!")
        skeys = list(cdict.keys())
        skeys.sort()
        outlist = []
        for sk in skeys:
            outlist.append("{} = {}\n".format(sk, cdict[sk]))

        with self.cfgfile.open('w') as fop:
            fop.writelines(outlist)

    def set_bool(self, key, value):
        """Sets boolean key in the settings file"""
        cdict = self.load()
        cdict[key.lower()] = bool(value)
        self.save(cdict)

    def set_int(self, key, value):
        """Sets integer key in the settings file"""
        cdict = self.load()
        cdict[key.lower()] = int(value)
        self.save(cdict)

    def set_path(self, wd, name=""):
        """Set the path in the settings file"""
        cdict = self.load()
        wdkey = "path {}".format(name.lower())
        cdict[wdkey] = wd
        self.save(cdict)

    def set_string(self, key, value=""):
        """Set the path in the settings file"""
        if value.strip():
            cdict = self.load()
            cdict[key] = value
            self.save(cdict)

    def set_string_list(self, key, value):
        """Set a list of strings"""
        if not isinstance(value, list):
            raise ValueError("`value` must be a list!")
        for ch in ["\t", "\n"]:
            if "".join(value).count(ch):
                raise ValueError("List items must not contain {}!".format(ch))
        cdict = self.load()
        cdict[key.lower()] = "\t".join(value)
        self.save(cdict)


class SettingsFileCache(SettingsFile):
    """A SettingsFile-based data cache"""

    def __init__(self, name):
        directory = appdirs.user_cache_dir()
        super(SettingsFileCache, self).__init__(name=name,
                                                defaults={},
                                                directory=directory)


class SettingsFileError(BaseException):
    pass



"""Test of data set functionalities"""
import pathlib

import numpy as np
from PyQt5 import QtWidgets
import pytest

import nanite.model as nmodel
import pyjibe.head


here = pathlib.Path(__file__).parent
jpkfile = here / "data" / "spot3-0192.jpk-force"


class MockModel():
    def __init__(self, model_key, **kwargs):
        # rebase on hertz model
        md = nmodel.models_available["hertz_para"]
        for key in dir(md):
            setattr(self, key, getattr(md, key))
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])
        self.model_key = model_key

    def __enter__(self):
        nmodel.register_model(self, self.__repr__())

    def __exit__(self, a, b, c):
        nmodel.models_available.pop(self.model_key)


def cleanup_autosave(jpkfile):
    """Remove autosave files"""
    path = jpkfile.parent
    files = path.glob("*.tsv")
    files = [f for f in files if f.name.startswith("pyjibe_")]
    [f.unlink() for f in files]


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    cleanup_autosave(jpkfile)
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    cleanup_autosave(jpkfile)


def test_ancillary_update_init(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {
            # take initial fit parameter of E
            "E": x.get_initial_fit_parameters(
                model_ancillaries=False)["E"].value},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=[jpkfile, jpkfile])
        war = main_window.subwindows[0].widget()
        # clear data
        war.tab_preprocess.list_preproc_applied.clear()
        war.cb_autosave.setChecked(0)
        # perform simple filter
        item = QtWidgets.QListWidgetItem()
        item.setText("compute_tip_position")
        war.tab_preprocess.list_preproc_applied.addItem(item)
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        war.on_tab_changed(1)
        assert len(war.data_set[0].preprocessing) == 1
        assert war.tab_preprocess.list_preproc_applied.count() == 1
        # The ancillary parameter gets its value from the default parameters
        assert atab.item(0, 1).text() == "3000"
        assert itab.item(0, 1).text() == "3000"
        # Now we change the initial parameter "E" and move on to the next
        # curve. Ancillary parameter "F" should also change.
        itab.item(0, 1).setText("2000")
        assert atab.item(0, 1).text() == "2000"
        it = war.list_curves.topLevelItem(1)
        war.list_curves.setCurrentItem(it)
        assert itab.item(0, 1).text() == "2000"
        assert atab.item(0, 1).text() == "2000"


def test_ancillary_update_nan(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {"E": np.nan},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=[jpkfile, jpkfile])
        war = main_window.subwindows[0].widget()
        # clear data
        war.tab_preprocess.list_preproc_applied.clear()
        war.cb_autosave.setChecked(0)
        # perform simple filter
        item = QtWidgets.QListWidgetItem()
        item.setText("compute_tip_position")
        war.tab_preprocess.list_preproc_applied.addItem(item)
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        assert atab.item(0, 1).text() == "nan"
        assert itab.item(0, 1).text() == "3000"


def test_ancillary_update_preproc_change(qtbot):
    with MockModel(
        compute_ancillaries=lambda x: {
            # i.e. model works only if there are multiple preproc steps
            "E": np.nan if len(x.preprocessing) == 1 else 2345},
        parameter_anc_keys=["E"],
        parameter_anc_names=["ancillary E guess"],
        parameter_anc_units=["Pa"],
            model_key="test1"):

        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=[jpkfile, jpkfile])
        war = main_window.subwindows[0].widget()
        # clear data
        war.tab_preprocess.list_preproc_applied.clear()
        war.cb_autosave.setChecked(0)
        # perform simple filter
        item = QtWidgets.QListWidgetItem()
        item.setText("compute_tip_position")
        war.tab_preprocess.list_preproc_applied.addItem(item)
        # disable weighting
        war.tab_fit.cb_weight_cp.setCheckState(0)
        # set mock model
        idx = war.tab_fit.cb_model.findData("test1")
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # perform fitting with standard parameters
        # set initial parameters in user interface
        itab = war.tab_fit.table_parameters_initial
        atab = war.tab_fit.table_parameters_anc
        war.on_tab_changed(1)
        assert len(war.data_set[0].preprocessing) == 1
        assert war.tab_preprocess.list_preproc_applied.count() == 1
        assert atab.item(0, 1).text() == "nan"
        assert itab.item(0, 1).text() == "3000"
        # up until here this is the same as `test_update_ancillary_nan`
        # now change preprocessing
        war.tabs.setCurrentIndex(0)  # actually switch tabs like a user
        item = QtWidgets.QListWidgetItem()
        item.setText("correct_tip_offset")
        war.tab_preprocess.list_preproc_applied.addItem(item)
        war.tabs.setCurrentIndex(1)  # triggers recomputation of anc
        assert len(war.data_set[0].preprocessing) == 2
        assert war.tab_preprocess.list_preproc_applied.count() == 2
        assert atab.item(0, 1).text() == "2345"
        assert itab.item(0, 1).text() == "2345"


def test_change_model_keep_parms(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile, jpkfile])
    war = main_window.subwindows[0].widget()
    # clear data
    war.tab_preprocess.list_preproc_applied.clear()
    war.cb_autosave.setChecked(0)
    # perform simple filter
    item = QtWidgets.QListWidgetItem()
    item.setText("compute_tip_position")
    war.tab_preprocess.list_preproc_applied.addItem(item)
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # set value for contact point
    itab.item(3, 1).setText(str(12345))
    # change the model to pyramidal
    pyr_name = nmodel.model_hertz_three_sided_pyramid.model_name
    pyr_idx = war.tab_fit.cb_model.findText(pyr_name)
    war.tab_fit.cb_model.setCurrentIndex(pyr_idx)
    # check that contact point is still the same
    assert float(itab.item(3, 1).text()) == 12345


def test_remember_initial_params(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile, jpkfile])
    war = main_window.subwindows[0].widget()
    # clear data
    war.tab_preprocess.list_preproc_applied.clear()
    war.cb_autosave.setChecked(0)
    # perform simple filter
    item = QtWidgets.QListWidgetItem()
    item.setText("compute_tip_position")
    war.tab_preprocess.list_preproc_applied.addItem(item)
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # disable weighting
    war.tab_fit.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    # change standard tip radius from 10 to 5
    assert float(itab.item(1, 1).text()) == 10
    itab.item(1, 1).setText(str(5))
    cl1 = war.list_curves.currentItem()
    cl2 = war.list_curves.itemBelow(cl1)
    war.list_curves.setCurrentItem(cl2)
    assert float(itab.item(1, 1).text()) == 5



import pkg_resources

import nanite.model as nmodel
import numpy as np
from PyQt5 import uic, QtCore, QtWidgets

from .. import units


class TabFit(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabFit, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_fit.ui")
        uic.loadUi(path_ui, self)

        # Setup the fitting tab
        id_para = 0
        # Model selection
        models_av = list(nmodel.models_available.keys())
        models_av.sort(key=lambda x: nmodel.models_available[x].model_name)
        for ii, key in enumerate(models_av):
            model = nmodel.models_available[key]
            self.cb_model.addItem(model.model_name, key)
            if key == "hertz_para":
                id_para = ii
        # Setz Hertz Parabolic as default
        self.cb_model.setCurrentIndex(id_para)
        # Axis selection
        self.cb_xaxis.addItem("tip position")
        self.cb_yaxis.addItem("force")

        # initial values, sources, drains for indentation depth
        self.indentation_depth_setup()

        # hide option to set right part of fitting range individually
        # (will be displayed when indentation depth is set individually,
        # i.e. `self.cb_delta_select.currentIndex == 1`)
        self.cb_right_individ.setVisible(False)

        # signals
        self.cb_segment.currentIndexChanged.connect(self.on_params_init)
        self.cb_xaxis.currentIndexChanged.connect(self.on_params_init)
        self.cb_yaxis.currentIndexChanged.connect(self.on_params_init)
        self.cb_model.currentIndexChanged.connect(self.on_model)
        self.cb_range_type.currentTextChanged.connect(self.on_params_init)
        self.table_parameters_anc.itemChanged.connect(self.on_params_anc)
        self.table_parameters_initial.itemChanged.connect(self.on_params_init)
        self.cb_weight_cp.stateChanged.connect(self.on_params_init)
        self.sp_weight_cp_um.valueChanged.connect(self.on_update_weights)
        self.sp_weight_cp_perc.valueChanged.connect(self.on_update_weights)

    @property
    def current_curve(self):
        return self.fd.current_curve

    @property
    def fd(self):
        return self.parent().parent().parent().parent()

    @property
    def fit_model(self):
        return nmodel.models_available[self.cb_model.currentData()]

    def anc_update_parameters(self, fdist):
        model_key = self.fit_model.model_key
        # Apply the current set of parameters
        # (some ancillary parameters depend on the correct initial parameters)
        fdist.fit_properties["params_initial"] = self.fit_parameters()
        # ancillaries
        anc = fdist.get_ancillary_parameters(model_key=model_key)
        anc_used = [ak for ak in anc if ak in self.fit_model.parameter_keys]
        if anc_used:
            self.widget_anc.setVisible(True)
            itab = self.table_parameters_initial
            atab = self.table_parameters_anc
            atab.blockSignals(True)
            rows_changed = self.assert_parameter_table_rows(atab,
                                                            len(anc_used),
                                                            cb_first=True,
                                                            read_only=True)
            row = 0
            for ii, ak in enumerate(self.fit_model.parameter_anc_keys):
                if ak not in anc_used:
                    continue
                # Get the human readable name of the parameter
                hrname = self.fit_model.parameter_anc_names[ii]
                # Determine unit scale, e.g. 1e6 [sic] for µm
                si_unit = self.fit_model.parameter_anc_units[ii]
                # Determine unit scale, e.g. 1e6 [sic] for µm
                scale = units.hrscale(hrname, si_unit=si_unit)
                label = units.hrscname(hrname, si_unit=si_unit)
                atab.verticalHeaderItem(row).setText(label)
                if rows_changed:
                    atab.item(row, 0).setCheckState(QtCore.Qt.Checked)
                atab.item(row, 1).setText("{:.5g}".format(anc[ak]*scale))
                # updates initial parameters if "use" is checked
                if atab.item(row, 0).checkState() == QtCore.Qt.Checked:
                    # update initial parameters
                    idx = self.fit_model.parameter_keys.index(ak)
                    text = atab.item(row, 1).text()
                    if text != "nan":
                        itab.item(idx, 1).setText(text)
                row += 1
            atab.blockSignals(False)
        else:
            self.widget_anc.setVisible(False)

    def assert_parameter_table_rows(self, table, rows, cb_first=False,
                                    read_only=False):
        """Make sure a QTableWidget has enough rows

        Parameters
        ----------
        table: instance of QtWidgets.QTableWidget
            The table that will be edited
        rows: int
            Number of rows
        cb_first: bool
            Set first column values as check-boxes.
            (This is used for fixing fit parameters in PyJibe)

        Returns
        -------
        rows_changed: bool
            Whether or not the number of rows changed
        """
        rows_changed = (table.rowCount() - rows) != 0
        table.setRowCount(rows)
        cols = table.columnCount()
        for rr in range(rows):
            hitem = QtWidgets.QTableWidgetItem()
            table.setVerticalHeaderItem(rr, hitem)
            for cc in range(cols):
                if table.item(rr, cc) is None:
                    item = QtWidgets.QTableWidgetItem()
                    table.setItem(rr, cc, item)
                else:
                    item = table.item(rr, cc)
                if cc == 0 and cb_first:
                    item.setFlags(QtCore.Qt.ItemIsUserCheckable
                                  | QtCore.Qt.ItemIsEnabled
                                  | QtCore.Qt.ItemIsEditable)
                elif read_only:
                    item.setFlags(QtCore.Qt.ItemIsEnabled)
                else:
                    item.setFlags(QtCore.Qt.ItemIsEnabled
                                  | QtCore.Qt.ItemIsEditable)

        return rows_changed

    def fit_approach_retract(self, fdist, update_ui=True):
        """Perform preprocessing and fit data

        Parameters
        ----------
        fdist: afmlib.indentaion.Indentaion
            Indentation curve to fit.
        update_ui: bool
            Update the user interface after fitting,
            i.e. displaying the results in
            `self.table_parameters_fitted`.
        """
        # segment
        segment = self.cb_segment.currentText().lower()
        # x axis
        x_axis = self.cb_xaxis.currentText()
        # y axis
        y_axis = self.cb_yaxis.currentText()
        # Get model key from dropdown list
        model_key = self.fit_model.model_key
        # Determine range type
        if self.cb_range_type.currentText() == "absolute":
            range_type = "absolute"
        else:
            range_type = "relative cp"
        # Determine range
        range_x = [self.sp_range_1.value() * units.scales["µ"],
                   self.sp_range_2.value() * units.scales["µ"]]
        # Remember range if applicable
        if self.cb_delta_select.currentIndex() == 1:
            self._indentation_depth_individual[fdist] = (
                self.sp_range_1.value(), self.sp_range_2.value())
        # Determine if we want to weight the contact point
        self.on_update_weights(on_params_init=False)
        if self.cb_weight_cp.checkState() == 2:
            weight_cp = self.sp_weight_cp_um.value() * units.scales["µ"]
        else:
            weight_cp = False
        # Determine if we want to autodetect the optimal indentation depth
        if self.cb_delta_select.currentIndex() == 2:
            optimal_fit_edelta = True
        else:
            optimal_fit_edelta = False
        # number of samples for edelta plot
        tab_edelta = self.fd.tab_edelta
        optimal_fit_num_samples = tab_edelta.sp_delta_num_samples.value()
        # fit parameters
        params = self.fit_parameters()
        # Perform fitting
        fdist.fit_model(model_key=model_key,
                        params_initial=params,
                        range_x=range_x,
                        range_type=range_type,
                        x_axis=x_axis,
                        y_axis=y_axis,
                        weight_cp=weight_cp,
                        segment=segment,
                        optimal_fit_edelta=optimal_fit_edelta,
                        optimal_fit_num_samples=optimal_fit_num_samples,
                        )
        ftab = self.table_parameters_fitted
        if fdist.fit_properties["success"]:
            # Perform automatic saving of results
            self.fd.autosave(fdist)
            # Display automatically detected optimal indentation depth
            if optimal_fit_edelta:
                # Set guessed indentation depth in GUI
                val = fdist.fit_properties["optimal_fit_delta"]
                self.sp_range_1.setValue(val/units.scales["µ"])
            if update_ui:
                # Display results in `self.table_parameters_fitted`
                fitpar = fdist.fit_properties["params_fitted"]
                # Display all varied parameters and expression parameters
                # (expression parameters are discouraged, but supported)
                fps = [p[1] for p in fitpar.items() if p[1].vary or p[1].expr]
                self.assert_parameter_table_rows(ftab, len(fps),
                                                 read_only=True)
                for ii, p in enumerate(fps):
                    # Get the human readable name of the parameter
                    idp = self.fit_model.parameter_keys.index(p.name)
                    hrname = self.fit_model.parameter_names[idp]
                    # SI unit
                    si_unit = self.fit_model.parameter_units[idp]
                    # Determine unit scale, e.g. 1e6 [sic] for µm
                    scale = units.hrscale(hrname, si_unit=si_unit)
                    label = units.hrscname(hrname, si_unit=si_unit)
                    ftab.verticalHeaderItem(ii).setText(label)
                    ftab.item(ii, 0).setText("{:.5g}".format(p.value*scale))
        else:
            if update_ui:
                inipar = fdist.fit_properties["params_initial"]
                fps = [p[1] for p in inipar.items() if p[1].vary]
                self.assert_parameter_table_rows(ftab, len(fps),
                                                 read_only=True)
                for ii, p in enumerate(fps):
                    # Get the human readable name of the parameter
                    idp = self.fit_model.parameter_keys.index(p.name)
                    hrname = self.fit_model.parameter_names[idp]
                    si_unit = self.fit_model.parameter_units[idp]
                    label = units.hrscname(hrname, si_unit=si_unit)
                    ftab.verticalHeaderItem(ii).setText(label)
                    ftab.item(ii, 0).setText("nan")

    def fit_parameters(self):
        """Returns parameters currently set in the GUI

        The parameter data is read out from
        `self.table_parameters_initial`. If the model that was
        previously set does not match the current model, only the
        parameters with the same name (existing in both models) are
        updated.
        """
        # Get model key from dropdown list
        model_key = self.fit_model.model_key
        # Get parameters from `self.table_parameters_initial`
        model = nmodel.models_available[model_key]
        params = model.get_parameter_defaults()
        itab = self.table_parameters_initial
        if itab.verticalHeaderItem(0):
            # Only update if there is already something set
            for ii, key in enumerate(list(params.keys())):
                p = params[key]
                if p.expr:
                    # Parameter has an expression, no update necessary
                    continue
                # Determine the scale for which we need to revert
                # the multiplications.
                # performed in `fit_update_parameters`.
                hrname = model.parameter_names[ii]
                si_unit = model.parameter_units[ii]
                scale = units.hrscale(hrname, si_unit=si_unit)
                for rr in range(itab.rowCount()):
                    # search for a row matching the parameter `p`
                    if itab.verticalHeaderItem(rr).text().startswith(hrname):
                        # update parameter `p`
                        state = itab.item(rr, 0).checkState()
                        if state == QtCore.Qt.Unchecked:
                            p.vary = True
                        else:
                            p.vary = False
                        p.set(float(itab.item(rr, 1).text())/scale)
                        p.min = float(itab.item(rr, 2).text())/scale
                        p.max = float(itab.item(rr, 3).text())/scale
                        break
        return params

    def fit_update_parameters(self, fdist):
        """Update the ancillary and initial parameters in the UI"""
        model_key = self.fit_model.model_key
        # set the model
        # - resets params_initial if model changed
        # - important for computing ancillary parameters
        fdist.fit_properties["model_key"] = model_key
        if fdist.fit_properties.get("params_initial", False):
            # (cannot coerce this into one line, because "params_initial"
            # can be None.)
            # set the parameters of the previous fit
            params = fdist.fit_properties["params_initial"]
        else:
            # use the initial model parameters
            params = self.fit_parameters()

        # parameter table
        itab = self.table_parameters_initial

        itab.setColumnWidth(0, 30)
        itab.setColumnWidth(1, 100)
        itab.setColumnWidth(2, 70)
        itab.setColumnWidth(3, 70)
        itab.blockSignals(True)

        self.assert_parameter_table_rows(itab, len(params), cb_first=True)
        params.update_constraints()  # in case we have expressions
        for ii, key in enumerate(list(params.keys())):
            p = params[key]
            # Get the human readable name of the parameter
            hrname = self.fit_model.parameter_names[ii]
            # SI unit
            si_unit = self.fit_model.parameter_units[ii]
            # Determine unit scale, e.g. 1e6 [sic] for µm
            scale = units.hrscale(hrname, si_unit=si_unit)
            label = units.hrscname(hrname, si_unit=si_unit)
            itab.verticalHeaderItem(ii).setText(label)
            if p.vary:
                state = QtCore.Qt.Unchecked
            else:
                state = QtCore.Qt.Checked
            itab.item(ii, 0).setCheckState(state)
            itab.item(ii, 1).setText("{:.5g}".format(p.value*scale))
            itab.item(ii, 2).setText(str(p.min*scale))
            itab.item(ii, 3).setText(str(p.max*scale))
            # grey out expression parameters
            if p.expr:
                for jj in range(4):
                    item = itab.item(ii, jj)
                    if jj == 0:  # check box
                        item.setCheckState(QtCore.Qt.Unchecked)
                    item.setFlags(QtCore.Qt.NoItemFlags)

        itab.blockSignals(False)

        # indentation depth
        if self.cb_delta_select.currentIndex() == 1:
            # Set indentation depth individually
            if fdist in self._indentation_depth_individual:
                left, right = self._indentation_depth_individual[fdist]
                self.fd.tab_edelta.delta_spin.setValue(left)
                if self.cb_right_individ.isChecked() and right is not None:
                    self.sp_range_2.setValue(right)

        # ancillaries
        self.anc_update_parameters(fdist)

    def indentation_depth_setup(self):
        """Initiate ranges (spin/slider) that allow inf values"""
        # Initialize individual indentation depth dictionary
        self._indentation_depth_individual = {}
        # Left
        self.sp_range_1.setValue(-np.inf)
        # Right
        self.sp_range_2.setValue(np.inf)
        # Signals
        self.cb_delta_select.currentIndexChanged['int'].connect(
            self.on_delta_select)

    def on_model(self):
        self.fd.on_model()

    def on_params_anc(self):
        self.fd.on_params_init()

    def on_params_init(self):
        self.fd.on_params_init()

    def on_delta_select(self, index):
        """The user selected a method for indentation depth determination

        This method is called when `tab_fit.cb_delta_select` or (indirectly)
        `tab_edelta.cb_delta_select` are changed.

        At the end, `on_params_init` is called, to redo the fit. After
        fitting, the indentation depth (`sp_range_1`) will be set by
        `fit_approach_retract`, if `value==2`.

        Parameters
        ----------
        index: int
            The selected index of the dropdown widget:
            - 0: Set indentation depth globally
            - 1: Set indentation depth individually
            - 2: Guess optimal indentation depth
        """
        # These widgets are disabled when the user wants
        # PyJibe to guess the indentation depth:
        global_disable = [self.sp_range_1]
        # These widgets are enabled when the user wants to
        # select an indentation depth for each curve:
        local_enable = [self.fd.tab_edelta.delta_label,
                        self.fd.tab_edelta.delta_spin,
                        self.fd.tab_edelta.delta_slider,
                        self.fd.tab_edelta.delta_btn_guess]

        if index == 0:
            # Set indentation depth globally
            [item.setEnabled(True) for item in global_disable]
            [item.setEnabled(False) for item in local_enable]
            self.cb_range_type.setEnabled(True)
        elif index == 1:
            # Set indentation depth individually
            [item.setEnabled(True) for item in global_disable]
            [item.setEnabled(True) for item in local_enable]
            self.cb_range_type.setEnabled(False)
            self.cb_range_type.setCurrentText("absolute")
            # Set/get indentation depth individually
            fdist = self.current_curve
            if fdist in self._indentation_depth_individual:
                left = self._indentation_depth_individual[fdist][0]
            else:
                left = self.fd.tab_edelta.delta_spin.value()
            self.fd.tab_edelta.delta_spin.setValue(left)
        elif index == 2:
            # Guess optimal indentation depth
            [item.setEnabled(False) for item in global_disable]
            [item.setEnabled(False) for item in local_enable]
            self.cb_range_type.setEnabled(False)
            self.cb_range_type.setCurrentText("absolute")
        else:
            raise ValueError("Invalid index '{}'!".format(index))

        # enable checkbox for right range
        if index == 1:
            self.cb_right_individ.setVisible(True)
        else:
            self.cb_right_individ.setVisible(False)
            self.cb_right_individ.setChecked(False)
        self.on_params_init()

    def on_update_weights(self, on_params_init=True):
        """Compute the weight in µm or % for user convenience

        This method updates the value of the interval range to
        suppress residuals near the contact point for the spin controls
        - `self.sp_weight_cp_um` (absolute [µm] radio button) and
        - `self.sp_weight_cp_perc` (% tip radius radio button),
        depending on which was updated by the user.

        Notes
        -----
        This method only updates the values if the model has a
        parameter "R", the radius of the indenter. In any other
        case, it will do nothing and disable the "% tip radius"
        radio button.
        """
        # First get the tip radius
        params = self.fit_model.get_parameter_defaults()
        itab = self.table_parameters_initial
        for ii, key in enumerate(list(params.keys())):
            if key == "R":
                radius = float(itab.item(ii, 1).text())
                # Make sure the "% tip radius" radio button is enabled,
                # because it could have been disabled for a fit model that
                # does not contain the parameter "R".
                self.rd_weigt_perc.setEnabled(True)
                # Did the user select "% tip radius"?
                percent = self.rd_weigt_perc.isChecked()

                if percent:
                    # set um value
                    perc = self.sp_weight_cp_perc.value()
                    self.sp_weight_cp_um.blockSignals(True)
                    self.sp_weight_cp_um.setValue(radius*perc/100)
                    self.sp_weight_cp_um.blockSignals(False)
                else:
                    # set percent value
                    val = self.sp_weight_cp_um.value()
                    self.sp_weight_cp_perc.blockSignals(True)
                    self.sp_weight_cp_perc.setValue(val/radius*100)
                    self.sp_weight_cp_perc.blockSignals(False)
                break
        else:
            # The key "R" could not be found. This means we
            # have a model that does not allow to compute the
            # indentation in percentage of the tip radius
            # (e.g. conical indenter).
            # We will force the user to select the "absolute [µm]"
            # radio button and disable the "% tip radius" radio button.
            self.rd_weight_um.setChecked(True)
            self.rd_weigt_perc.setEnabled(False)

        if on_params_init:
            self.on_params_init()



"""Helper methods for handling units in PyJibe"""


def si2hr(name, value, si_unit=None):
    """Convert an SI unit to a human readable unit

    Parameters
    ----------
    name: str
        The parameter name string, see `human_units`.
    value: float or array
        The value of the parameters that will be converted
        to a human readable scale.

    Returns
    -------
    hr_value: float
        The human readable value.
    scaleunit: str
        The unit including the scale, e.g. "µm".
    """
    name = name.lower()
    if name in human_units:
        scalename, unit = human_units[name]
        scale = scales[scalename]
        hr_value = value/scale
        scaleunit = scalename + unit
    elif si_unit in default_scales:
        scalename = default_scales[si_unit]
        scale = scales[scalename]
        hr_value = value/scale
        scaleunit = scalename + si_unit
    elif si_unit is not None:
        hr_value = value
        scaleunit = si_unit
    else:
        hr_value = value
        scaleunit = ""
    return hr_value, scaleunit


def hrscale(name, si_unit=None):
    """Returns the multiplier for unit scale conversion, e.g. 1e6 for µm"""
    name = name.lower()
    if name in human_units:
        scalename, _unit = human_units[name]
        scale = 1/scales[scalename]
    elif si_unit in default_scales:
        scalename = default_scales[si_unit]
        scale = 1/scales[scalename]
    else:
        scale = 1
    return scale


def hrunit(name, si_unit=None):
    """Returns the unit name for scale conversion, e.g. µm"""
    name = name.lower()
    if name in human_units:
        scalename, unit = human_units[name]
        scaleunit = scalename + unit
        if si_unit is not None and si_unit != unit:
            raise ValueError(
                "Bad `si_unit` '{}' given for '{}', expected '{}'! ".format(
                    si_unit, name, unit))
    elif si_unit in default_scales:
        scaleunit = default_scales[si_unit] + si_unit
    elif si_unit is not None:
        scaleunit = si_unit
    else:
        scaleunit = ""
    return scaleunit


def hrscname(name, si_unit=None):
    """Returns the name with human readable units, e.g. Force [nN]"""
    unit = hrunit(name, si_unit=si_unit)
    if unit == "":
        hrscname = name
    else:
        hrscname = "{} [{}]".format(name, unit)
    return hrscname


# TODO:
# - make the scales user-editable
default_scales = {"deg": "",
                  "m": "µ",
                  "m/s": "µ",
                  "m/V": "n",
                  "N": "n",
                  "Pa": "",
                  "Pa·s": "",
                  }

human_units = {}
human_units["young's modulus"] = [default_scales["Pa"], "Pa"]
human_units["tip radius"] = ["µ", "m"]
human_units["contact point"] = ["n", "m"]
human_units["force"] = [default_scales["N"], "N"]
human_units["force baseline"] = ["p", "N"]
human_units["tip position"] = [default_scales["m"], "m"]

scales = {}
scales["k"] = 1e3
scales[""] = 1
scales["m"] = 1e-3
scales["µ"] = 1e-6
scales["n"] = 1e-9
scales["p"] = 1e-12



import codecs
import hashlib
import pathlib
import pkg_resources

import afmformats
import h5py
from PyQt5 import uic, QtWidgets


class ConvertDialog(QtWidgets.QDialog):
    _instance_counter = 0

    def __init__(self, parent, *args, **kwargs):
        """Data conversion dialog"""
        super(ConvertDialog, self).__init__(parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.head",
                                                  "dlg_tool_convert.ui")
        uic.loadUi(path_ui, self)

        self._file_list = []

        self.toolButton_browse.clicked.connect(self.on_browse)
        self.toolButton_clear.clicked.connect(self._file_list.clear)

    def _convert_merge(self):
        file, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.parent(),
            "Output file",
            "",
            "HDF5 file (*.h5)")
        if not file:
            return False
        if not file.endswith(".h5"):
            file += ".h5"
        # HDF5 format
        with h5py.File(file, "w") as h5:
            for path in self.file_list:
                fdlist = afmformats.load_data(path)
                for fdist in fdlist:
                    fdist.export(h5,
                                 metadata=self.get_metadata_keys(fdist),
                                 fmt="hdf5")
        return True

    def _convert_curve(self):
        out_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self.parent(), "Select output directory", "")
        if out_dir:
            for path in self.file_list:
                path = pathlib.Path(path)
                epath = codecs.encode(str(path), encoding="utf-8",
                                      errors="ignore")
                stem = path.name + "_" + hashlib.md5(epath).hexdigest()[:5]
                fdlist = afmformats.load_data(path)
                for fdist in fdlist:
                    name = "{}_{}.{}".format(stem, fdist.enum, self.format)
                    opath = pathlib.Path(out_dir) / name
                    fdist.export(opath,
                                 metadata=self.get_metadata_keys(fdist),
                                 fmt=self.format)
            return True
        else:
            return False  # do not close the dialog

    def _convert_unaltered(self):
        out_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self.parent(), "Select output directory", "")
        if out_dir:
            for path in self.file_list:
                path = pathlib.Path(path)
                epath = codecs.encode(str(path), encoding="utf-8",
                                      errors="ignore")
                name = "{}_{}.{}".format(path.name,
                                         hashlib.md5(epath).hexdigest()[:5],
                                         self.format)
                opath = pathlib.Path(out_dir) / name

                with h5py.File(opath, mode="w") as h5:
                    fdlist = afmformats.load_data(path)
                    for fdist in fdlist:
                        fdist.export(h5,
                                     metadata=self.get_metadata_keys(fdist),
                                     fmt=self.format)
            return True
        else:
            return False  # do not close the dialog

    @property
    def file_list(self):
        return sorted(self._file_list)

    @property
    def format(self):
        # file format
        if self.radioButton_hdf5.isChecked():
            fmt = "h5"
        else:
            fmt = "tab"
        return fmt

    def add_file(self, path):
        self.listWidget.addItem(path)
        self._file_list.append(path)

    def convert(self):
        # file grouping
        if self.radioButton_unaltered.isChecked():
            return self._convert_unaltered()
        elif self.radioButton_merge.isChecked():
            return self._convert_merge()
        else:
            return self._convert_curve()

    def done(self, r):
        if r:
            if self.listWidget.count():
                if not self.convert():
                    return
            else:
                QtWidgets.QMessageBox.information(
                    self, "No selection", "You did not select any data!")
                return  # do not close the dialog

        super(ConvertDialog, self).done(r)

    def dragEnterEvent(self, e):
        """Whether files are accepted"""
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        """Add dropped files to view"""
        urls = e.mimeData().urls()
        pathlist = []
        for ff in urls:
            pp = pathlib.Path(ff.toLocalFile())
            if pp.is_dir():
                for ext in afmformats.supported_extensions:
                    pathlist += list(pp.rglob("*" + ext))
            elif pp.suffix in afmformats.supported_extensions:
                pathlist.append(pp)
        for pp in pathlist:
            self.add_file(str(pp))

    def get_metadata_keys(self, fdist):
        metadata = fdist.metadata
        if not self.checkBox_storage.isChecked():
            for key in metadata.get_summary()["storage"]:
                if key in metadata:
                    metadata.pop(key)
        return list(metadata.keys())

    def on_browse(self):
        ext = " ".join(["*"+e for e in afmformats.supported_extensions])
        files, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self.parent(),
            "Select data files",
            "",
            "All supported files ({})".format(ext))

        for ff in files:
            self.add_file(ff)



"""Test of data set functionalities"""
import pathlib

import lmfit
import numpy as np
from PyQt5 import QtWidgets
import pytest

import nanite.model
from nanite.model import weight
import pyjibe.head


here = pathlib.Path(__file__).parent
jpkfile = here / "data" / "spot3-0192.jpk-force"


class MockModelExpr():
    def __init__(self, **kwargs):
        """E and E1 add up to the actual emodulus. E1 is varied indirectly"""
        self.model_doc = """Mock model with constraint"""
        self.model_func = MockModelExpr.hertz_constraint
        self.model_key = "hertz_constraint"
        self.model_name = "Hertz with constraint "
        self.parameter_keys = ["E", "R", "nu", "virtual_parameter",
                               "E1", "contact_point", "baseline"]
        self.parameter_names = ["Young's Modulus", "Tip Radius",
                                "Poisson's Ratio", "Virtual Parameter",
                                "Another Modulus", "Contact Point",
                                "Force Baseline"]
        self.parameter_units = ["Pa", "m", "", "Pa", "Pa", "m", "N"]
        self.valid_axes_x = ["tip position"]
        self.valid_axes_y = ["force"]

    @staticmethod
    def get_parameter_defaults():
        # The order of the parameters must match the order
        # of ´parameter_names´ and ´parameter_keys´.
        params = lmfit.Parameters()
        params.add("E", value=1e3, min=0, vary=False)
        params.add("R", value=10e-6, vary=False)
        params.add("nu", value=.5, vary=False)
        params.add("virtual_parameter", value=10, min=0, vary=True)
        params.add("E1", expr="virtual_parameter+E")
        params.add("contact_point", value=0)
        params.add("baseline", value=0)
        return params

    @staticmethod
    def hertz_constraint(delta, E, R, nu, virtual_parameter, E1,
                         contact_point=0, baseline=0):
        aa1 = 4 / 3 * E1 / (1 - nu ** 2) * np.sqrt(R)

        root = contact_point - delta
        pos = root > 0
        bb = np.zeros_like(delta)
        bb[pos] = (root[pos]) ** (3 / 2)
        cc = np.zeros_like(delta)
        cc[pos] = 1 - 0.15 * root[pos] / R
        return aa1 * bb * cc + baseline

    @staticmethod
    def model(params, x):
        if x[0] < x[-1]:
            revert = True
        else:
            revert = False
        if revert:
            x = x[::-1]
        mf = MockModelExpr.hertz_constraint(
            E=params["E"].value,
            delta=x,
            R=params["R"].value,
            nu=params["nu"].value,
            virtual_parameter=params["virtual_parameter"].value,
            E1=params["E1"].value,
            contact_point=params["contact_point"].value,
            baseline=params["baseline"].value)
        if revert:
            return mf[::-1]
        return mf

    @staticmethod
    def residual(params, delta, force, weight_cp=5e-7):
        """ Compute residuals for fitting

        Parameters
        ----------
        params: lmfit.Parameters
            The fitting parameters for `model`
        delta: 1D ndarray of lenght M
            The indentation distances
        force: 1D ndarray of length M
            The corresponding force data
        weight_cp: positive float or zero/False
            The distance from the contact point until which
            linear weights will be applied. Set to zero to
            disable weighting.
        """
        md = MockModelExpr.model(params, delta)
        resid = force - md

        if weight_cp:
            # weight the curve so that the data around the contact_point do
            # not affect the fit so much.
            weights = weight.weight_cp(
                cp=params["contact_point"].value,
                delta=delta,
                weight_dist=weight_cp)
            resid *= weights
        return resid

    def __enter__(self):
        nanite.model.register_model(self, self.__repr__())
        return self

    def __exit__(self, a, b, c):
        nanite.model.models_available.pop(self.model_key)


def cleanup_autosave(jpkfile):
    """Remove autosave files"""
    path = jpkfile.parent
    files = path.glob("*.tsv")
    files = [f for f in files if f.name.startswith("pyjibe_")]
    [f.unlink() for f in files]


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    cleanup_autosave(jpkfile)
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    cleanup_autosave(jpkfile)


def test_model_expr_base(qtbot):
    """Expressions sanity check

    This test is just to make sure that parameters that have an
    expression are not overridden by the UI.
    """
    with MockModelExpr() as mod:
        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=[jpkfile, jpkfile])
        war = main_window.subwindows[0].widget()
        # clear data
        war.tab_preprocess.list_preproc_applied.clear()
        war.cb_autosave.setChecked(0)
        # perform simple filter
        item = QtWidgets.QListWidgetItem()
        item.setText("compute_tip_position")
        war.tab_preprocess.list_preproc_applied.addItem(item)
        # set mock model
        idx = war.tab_fit.cb_model.findData(mod.model_key)
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # Now there should be fit results
        itab = war.tab_fit.table_parameters_initial
        # set better value for contact point
        itab.item(5, 1).setText(str(18000))
        fdist = war.data_set[0]
        parmsi = fdist.fit_properties["params_initial"]
        parmsf = fdist.fit_properties["params_fitted"]
        assert parmsi["E1"].expr == "virtual_parameter+E"
        assert parmsf["E1"].expr == "virtual_parameter+E"
        assert np.allclose(parmsf["E1"].value, 15388.14166)


if __name__ == "__main__":
    with MockModelExpr() as mod:
        from pyjibe.__main__ import main
        main()



import numpy as np

from PyQt5 import QtWidgets, QtGui


class InfDoubleSpinBox(QtWidgets.QDoubleSpinBox):
    def __init__(self, *args, **kwargs):
        super(InfDoubleSpinBox, self).__init__(*args, **kwargs)
        self.setMinimum(-np.inf)
        self.setMaximum(np.inf)
        self.validator = FloatValidator()
        self.setDecimals(1000)
        self.setSingleStep(0.1)

    def validate(self, text, position):
        return self.validator.validate(text, position)

    def fixup(self, text):
        return self.validator.fixup(text)

    def valueFromText(self, text):
        return convert_string_to_float(text)

    def textFromValue(self, value):
        return str(value)


class FloatValidator(QtGui.QValidator):
    def validate(self, string, position):
        if valid_float_string(string):
            return self.Acceptable, string, position
        if string == "" or string[position-1] in 'e.-+':
            return self.Intermediate, string, position
        return self.Invalid, string, position

    def fixup(self, text):
        try:
            val = convert_string_to_float(text)
        except ValueError:
            val = ""
        return str(val)


def valid_float_string(string):
    try:
        convert_string_to_float(string)
    except ValueError:
        valid = False
    else:
        valid = True
    return valid


def convert_string_to_float(string):
    try:
        val = float(string)
    except ValueError:
        string = string.strip()
        if string.startswith("-"):
            asign = -1
        else:
            asign = 1
        string = string.strip("+-")
        for iid in ["i", "in", "inf"]:
            if string == iid:
                val = asign*np.inf
                break
        else:
            raise ValueError("Not a valid float!")
    return val



"""Test of data set functionalities"""
import pathlib

from PyQt5 import QtWidgets, QtCore
import pytest


import pyjibe.head


here = pathlib.Path(__file__).parent
jpkfile = here / "data" / "map2x2_extracted.jpk-force-map"


def cleanup_autosave(jpkfile):
    """Remove autosave files"""
    path = jpkfile.parent
    files = path.glob("*.tsv")
    files = [f for f in files if f.name.startswith("pyjibe_")]
    [f.unlink() for f in files]


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    cleanup_autosave(jpkfile)
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    cleanup_autosave(jpkfile)


def test_qmap_with_unused_curves(qtbot):
    """Uncheck "use" in the curve list and switch to the qmap tab"""
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=[jpkfile])
    war = main_window.subwindows[0].widget()
    # uncheck first curve
    cl1 = war.list_curves.currentItem()
    cl1.setCheckState(3, QtCore.Qt.Unchecked)
    war.tabs.setCurrentIndex(5)
    QtWidgets.QApplication.processEvents()



from collections import OrderedDict
import codecs
import numbers

import numpy as np

from afmformats import meta
import nanite.model as nmodel

from .. import units

#: Valid export choices in `save_tsv_metadata_results`
EXPORT_CHOICES = list(meta.META_FIELDS.keys()) + [
    "params_ancillary",
    "params_fitted",
    "params_initial",
    "rating"]


def save_tsv_metadata_results(filename, fdist_list, which=EXPORT_CHOICES):
    """Export metadata and fitting parameters

    Parameters
    ----------
    filename: str
        Path to store data to
    fdist_list: list-like
        List of :class:`nanite.Indentation` instances
    which: list of str
        Valid for choices to export (see :data:`EXPORT_CHOICES`)
    """
    if np.sum([k not in EXPORT_CHOICES for k in which]):
        raise ValueError("Found invalid export choices.")

    columns = OrderedDict()
    # Metadata
    for ii, fdist in enumerate(fdist_list):
        meta = fdist.metadata.get_summary()
        for topic in meta:
            if topic in which:
                for kk in meta[topic]:
                    label, hrvalue = get_unitname_value_meta(
                        key=kk, value=meta[topic][kk])
                    if label not in columns:
                        columns[label] = [np.nan] * len(fdist_list)
                    columns[label][ii] = hrvalue

        # Parameters
        if "model_key" in fdist.fit_properties:
            model_key = fdist.fit_properties["model_key"]

            # Ancillary
            if "params_ancillary" in which:
                anc_dict = fdist.get_ancillary_parameters()
                for kk in anc_dict:
                    label, hrvalue = get_unitname_value(
                        name=nmodel.get_parm_name(model_key, kk),
                        value=anc_dict[kk],
                        unit=nmodel.get_parm_unit(model_key, kk))
                    if label not in columns:
                        columns[label] = [np.nan] * len(fdist_list)
                    columns[label][ii] = hrvalue

            # Initial
            if "params_initial" in which:
                if "params_initial" in fdist.fit_properties:
                    fp = fdist.fit_properties["params_initial"]
                    for ki in fp:
                        if not fp[ki].vary:
                            label, hrvalue = get_unitname_value(
                                name=nmodel.get_parm_name(model_key, ki),
                                value=fp[ki].value,
                                unit=nmodel.get_parm_unit(model_key, ki))
                            if label not in columns:
                                columns[label] = [np.nan] * len(fdist_list)
                            columns[label][ii] = hrvalue

            # Fitted
            if "params_fitted" in which:
                if "params_fitted" in fdist.fit_properties:
                    fp = fdist.fit_properties["params_fitted"]
                    for ki in fp:
                        if fp[ki].vary:
                            label, hrvalue = get_unitname_value(
                                name=nmodel.get_parm_name(model_key, ki),
                                value=fp[ki].value,
                                unit=nmodel.get_parm_unit(model_key, ki))
                            if label not in columns:
                                columns[label] = [np.nan] * len(fdist_list)
                            columns[label][ii] = hrvalue

                    # Additional fit parameters
                    props = {"xmin": ("Fit interval minimum", "m"),
                             "xmax": ("Fit interval maximum", "m"),
                             "segment": ("Fit segment", ""),
                             "weight_cp": ("Fit weight contact point", "m"),
                             "model_key": ("Fit model", ""),
                             }
                    for prop in props:
                        label, hrvalue = get_unitname_value(
                            name=props[prop][0],
                            value=fdist.fit_properties[prop],
                            unit=props[prop][1])
                        if label not in columns:
                            columns[label] = [np.nan] * len(fdist_list)
                        columns[label][ii] = hrvalue

            if "rating" in which:
                rdict = fdist.get_rating_parameters()
                for label in ["Regressor", "Training set", "Rating"]:
                    if label not in columns:
                        columns[label] = [np.nan] * len(fdist_list)
                    columns[label][ii] = rdict[label]

    save_tsv(filename, columns)


def save_tsv(filename, column_dict):
    """Export data set to a .tsv file

    Parameters
    ----------
    filename: str
        Filename to save as
    column_dict: dict of lists
        List with column headers and content to save.
        E.g.: {"Column One": [1.1, 2.2, 3.3, 4.4],
               "Column Two": [nan, 2.0, 4.0, 1.0],
               }
    """
    with codecs.open(filename, "wb") as fd:
        fd.write(codecs.BOM_UTF8)
    with codecs.open(filename, "a", encoding="utf-8") as fd:
        # Write header:
        header = "\t".join([d for d in list(column_dict.keys())])
        fd.write(header+"\r\n")

        # Write data
        cols = list(column_dict.values())
        # Transpose data
        data = transpose_list(cols)
        # Save line-wise
        for d in data:
            line = "\t".join([format_content(it) for it in d])
            fd.write(line+"\r\n")


def format_content(value):
    if isinstance(value, numbers.Number):
        return "{:.5g}".format(value)
    else:
        return "{}".format(value)


def get_unitname_value_meta(key, value):
    """Return header / value pair for tsv export of Indentation metadata"""
    name, unit, validator = meta.DEF_ALL[key]
    if isinstance(value, numbers.Number):
        if not np.isnan(value):
            value = validator(value)
    else:
        value = validator(value)
    return get_unitname_value(name, value, unit)


def get_unitname_value(name, value, unit):
    """Return header / value pair for tsv export"""
    if isinstance(value, numbers.Number):
        hrvalue, scunit = units.si2hr(name=name, value=value, si_unit=unit)
        strunit = " [{}]".format(scunit) if scunit else ""
        header = name + strunit
    else:
        hrvalue = value
        header = name
    return header, hrvalue


def transpose_list(m):
    height = len(m)
    width = len(m[0])
    tr = [[m[row][col] for row in range(0, height)] for col in range(0, width)]
    return tr



def main(splash=True):
    import os
    import pkg_resources
    import sys

    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    # Note:
    # Having the image file *not* in a submodule of PyJibe
    # seems to cause the splash to display earlier, because
    # presumably `pkg_resources` internally imports modules.
    imdir = pkg_resources.resource_filename("pyjibe", "img")

    if splash:
        from PyQt5.QtWidgets import QSplashScreen
        from PyQt5.QtGui import QPixmap
        splash_path = os.path.join(imdir, "splash.png")
        splash_pix = QPixmap(splash_path)
        splash = QSplashScreen(splash_pix)
        splash.setMask(splash_pix.mask())
        splash.show()
        # make sure Qt really displays the splash screen
        app.processEvents()

    from PyQt5 import QtCore, QtGui
    from .head import PyJibe

    # Set Application Icon
    icon_path = os.path.join(imdir, "icon.png")
    app.setWindowIcon(QtGui.QIcon(icon_path))

    # Use dots as decimal separators
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.C))

    window = PyJibe()

    if splash:
        splash.finish(window)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import numbers
import pkg_resources

from afmformats import meta
from nanite import model
import numpy as np
from PyQt5 import uic, QtWidgets

from .. import units


class TabInfo(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(TabInfo, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("pyjibe.fd",
                                                  "tab_info.ui")
        uic.loadUi(path_ui, self)

    def update_info(self, fdist):
        hr_info = {}

        # Metadata
        msum = fdist.metadata.get_summary()
        for sec in msum:
            if sec == "qmap":
                if msum["qmap"] and not has_all_nans(msum["qmap"]):
                    textq = []
                    for kk in msum["qmap"]:
                        textq.append(get_string_rep_meta(kk, msum["qmap"][kk]))
                    hr_info["QMap"] = textq
            else:
                atext = []
                for kk in msum[sec]:
                    atext.append(get_string_rep_meta(kk, msum[sec][kk]))
                hr_info[sec.capitalize()] = atext

        # Ancillaries
        anc_dict = fdist.get_ancillary_parameters()
        if anc_dict:
            text_meta = []
            model_key = fdist.fit_properties["model_key"]
            for kk in anc_dict:
                text_meta.append(
                    get_string_rep(name=model.get_parm_name(model_key, kk),
                                   value=anc_dict[kk],
                                   unit=model.get_parm_unit(model_key, kk)))
            hr_info["Ancillaries"] = text_meta

        text = []
        for sec in sorted(hr_info.keys()):
            text.append("<b>{}:</b>".format(sec))
            text += sorted(hr_info[sec])
            text.append("")

        textstring = "<br>".join(text)

        # remember the scroll position
        vval = self.info_text.verticalScrollBar().value()
        hval = self.info_text.horizontalScrollBar().value()
        self.info_text.setText(textstring)
        self.info_text.verticalScrollBar().setValue(vval)
        self.info_text.horizontalScrollBar().setValue(hval)


def get_string_rep_meta(key, value):
    """Return a nice string representation for a key in metadata"""
    name, unit, validator = meta.DEF_ALL[key]
    if isinstance(value, numbers.Number):
        if not np.isnan(value):
            value = validator(value)
    else:
        value = validator(value)
    return get_string_rep(name, value, unit)


def get_string_rep(name, value, unit):
    """Return pretty-formatted string for key, value, and unit"""

    if isinstance(value, numbers.Number):
        hrvalue, scunit = units.si2hr(name=name, value=value, si_unit=unit)

        if np.isnan(value):
            rep = "<span style='color:#757575'>{}: unknown</span>".format(name)
        else:
            rep = "{}: {:.5g}{}".format(name,
                                        hrvalue,
                                        " " + scunit if scunit else "",
                                        )
    else:
        rep = "{}: {}".format(name, value)

    return rep


def has_all_nans(adict):
    """Check whether a dictionary as all-nan values"""
    allnan = True
    for value in adict.values():
        if isinstance(value, numbers.Number):
            if not np.isnan(value):
                allnan = False
                break
        else:
            allnan = False
            break
    return allnan



import os
import pathlib
import pkg_resources

from PyQt5 import QtCore, QtGui, QtWidgets

from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
from matplotlib import cbook


class FileDialog(QtWidgets.QFileDialog):
    def __init__(self, *args):
        """A dialog that lets the user select multiple directories"""
        QtWidgets.QFileDialog.__init__(self, *args)
        self.setOption(self.DontUseNativeDialog, True)
        self.setFileMode(self.DirectoryOnly)

        self.tree = self.findChild(QtWidgets.QTreeView)
        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)

        self.list = self.findChild(QtWidgets.QListView)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)

        for view in self.findChildren((QtWidgets.QListView,
                                       QtWidgets.QTreeView)):
            if isinstance(view.model(), QtWidgets.QFileSystemModel):
                view.setSelectionMode(
                    QtWidgets.QAbstractItemView.MultiSelection)

    def getDirectory(self):
        dirs = self.selectedFiles()
        if dirs:
            path = pathlib.Path(dirs[0])
            if len(dirs) > 1:
                path = path.parent
        return str(path)

    def selectedFilesRecursive(self):
        dirs = self.selectedFiles()
        files = []
        for d in dirs:
            for rt, _d, fs in os.walk(d):
                for f in fs:
                    files.append(os.path.join(rt, f))
        return files


# TODO:
# - use the capabilities of the new toolbars?

class NavigationToolbarCustom(NavigationToolbar2QT):
    """A custom toolbar that allows other icons"""

    def __init__(self, *args, **kwargs):
        super(NavigationToolbarCustom, self).__init__(*args, **kwargs)

    def _icon(self, name, color=None):
        """Override matplotlibs `_icon` function to get custom icons"""
        name = name.replace('.png', '_large.png')
        impath = str(cbook._get_data_path('images', name))
        if not os.path.exists(impath):
            imdir = pkg_resources.resource_filename("pyjibe", "img")
            impath = os.path.join(imdir, name)
        pm = QtGui.QPixmap(impath)
        pm.setDevicePixelRatio(self.devicePixelRatioF() or 1)
        if self.palette().color(self.backgroundRole()).value() < 128:
            icon_color = self.palette().color(self.foregroundRole())
            mask = pm.createMaskFromColor(QtGui.QColor('black'),
                                          QtCore.Qt.MaskOutColor)
            pm.fill(icon_color)
            pm.setMask(mask)
        return QtGui.QIcon(pm)

    def save_data(self, *args):
        fname, _e = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save raw data", "", "Tab Separated Values (*.tab)")
        if fname:
            if not fname.endswith(".tab"):
                fname += ".tab"
            self.save_data_callback(fname)

    def save_data_callback(self, filename):
        """Save the plot data to a tab file
        """
        raise ValueError("This Method must be overridden!")


class NavigationToolbarIndent(NavigationToolbarCustom):
    def __init__(self, *args, **kwargs):
        self.toolitems = (
            ('Home', 'Reset plot', 'home', 'home'),
            ('Pan', 'Pan tool', 'move', 'pan'),
            ('Zoom', 'Zoom tool', 'zoom_to_rect', 'zoom'),
            ('Save', 'Save approach-retract curve image',
             'saveimg', 'save_figure'),
            ('Save', 'Save approach-retract curve data',
             'savedat', 'save_data'),
        )
        super(NavigationToolbarIndent, self).__init__(*args, **kwargs)


class NavigationToolbarEDelta(NavigationToolbarCustom):
    def __init__(self, *args, **kwargs):
        self.toolitems = (
            ('Home', 'Reset plot', 'home', 'home'),
            ('Pan', 'Pan tool', 'move', 'pan'),
            ('Zoom', 'Zoom tool', 'zoom_to_rect', 'zoom'),
            ('Save', 'Save E(δ) curve image', 'saveimg', 'save_figure'),
            ('Save', 'Save E(δ) data', 'savedat', 'save_data'),
        )
        super(NavigationToolbarEDelta, self).__init__(*args, **kwargs)


class NavigationToolbarQMap(NavigationToolbarCustom):
    def __init__(self, *args, **kwargs):
        self.toolitems = (
            ('Home', 'Reset plot', 'home', 'home'),
            ('Pan', 'Pan tool', 'move', 'pan'),
            ('Zoom', 'Zoom tool', 'zoom_to_rect', 'zoom'),
            ('Save', 'Save quantitative map image', 'saveimg', 'save_figure'),
            ('Save', 'Save quantitative map data', 'savedat', 'save_data'),
        )
        super(NavigationToolbarQMap, self).__init__(*args, **kwargs)
